% -------------------------------------------------------------------------
%                     How to prestress compliant mechanisms
%                      for a targeted stiffness adjustment
%
%                       Kristian Mauser, Alexander Hasse
%                   Journal of Smart Materials and Structures
% -------------------------------------------------------------------------

% *************************** Disclaimer *********************************
% The authors reserve all rights but do not guarantee that the code is
% free from errors. In addition, they are in no case liable for any event 
% caused by the use of the program.  
% ************************************************************************

% *************************** Input File *********************************
% nodes ........... node coordinates in column 2 and 3
% elements ........ beam elements and corresponding nodes
% nElements ....... total number of elements
% nNodes .......... total number of nodes
% f_xi ............ predefined load cases (compare Eq. 24); each column 
%                   represents one load case; rows reflect the degrees of 
%                   freedom (DoFs)
% k ............... total number of load cases
% t ............... thickness t of the beam elements
% Ke .............. material stiffness matrix
% cellKg .......... initial stress stiffness matrices Kg_xi, with 
%                   xi = 1,...k, for all k load cases f_xi
% freeDoFs ........ unconstrainted DoFs
% allDoFs ......... all DoFs (column 1) and corresponding nodes (column 2)

clear, clf, close

% load structural input parameters
load('Cartwheelhinge_Structur.mat')

% calculation of the eigenvectors of Ke
[phi, lambda] = eig(Ke(freeDoFs,freeDoFs));

% sort in ascending order
[lambda, Index] = sort(diag(lambda));
phi = phi(:,Index);

% subdivision into primary stiffness and secondary stiffness
% ignore three rigid-body modes, as there are no constraints
% 4th displacement mode is the primary stiffness mode
i = 4;

% 5th displacement mode is the secondary stiffness mode taken into account
% --> nu = 1
j = 5;

% lambda_circ_e_i (Eq. 20)
lambda_circ_e_i = phi(:,i)' * Ke(freeDoFs,freeDoFs) * phi(:,i);
    
% lambda_circ_e_j (Eq. 22)
lambda_circ_e_j = phi(:,j)' * Ke(freeDoFs,freeDoFs) * phi(:,j);

% lambda_circ_g_i_xi (Eq. 20) and lambda_circ_g_j_xi (Eq. 22) for each load
% case
lambda_circ_g_i_xi = zeros(k,1);
lambda_circ_g_j_xi = zeros(k,1);

for m = 1:k
    lambda_circ_g_i_xi(m,1) = phi(:,i)' * cellKg{1,m}(freeDoFs,freeDoFs) * phi(:,i);
    lambda_circ_g_j_xi(m,1) = phi(:,j)' * cellKg{1,m}(freeDoFs,freeDoFs) * phi(:,j);
end

% desired lambda value
lambda_d = 0.5 * lambda(i,1);


% *************************** Optimization ********************************
cvx_begin
cvx_precision best

    variable x(k)

    % Objective function
    minimize(sum(abs(x)))
    
    subject to

        % calculate f (Eq. 16)
        for n = 1:k
            x_xi_f_xi(:,n) = x(n,1)' .* f_xi(freeDoFs,n);
        end

        f = sum(x_xi_f_xi,2);
        lambda_circ_i = lambda_circ_e_i + sum(x .* lambda_circ_g_i_xi(1:end,1));
        lambda_circ_j = lambda_circ_e_j + sum(x .* lambda_circ_g_j_xi(1:end,1));

        % constraints
        lambda_d - lambda_circ_i == 0;
        phi(:,i)' * f == 0;
        phi(:,j)' * f == 0;
        -lambda_circ_j * lambda_circ_e_i + lambda_circ_e_j * lambda_circ_i <= 0;
           
cvx_end
clearvars x_xi_f_xi

% calculate resultant prestressing forces (Eq. 16)
F = sum(x' .* f_xi,2);

% for overview purposes, not necessary
F(abs(F) < 1e-8) = 0;


% *************************** Visualization *******************************
% Plot structure
for l = 1:nElements
    plotx = [nodes(elements(l,2),2), nodes(elements(l,3),2)];
    ploty = [nodes(elements(l,2),3), nodes(elements(l,3),3)];
    h(1)  = plot(plotx, ploty, '-k', 'LineWidth', t(l)');
    hold on
end

% Plot load cases
Freshape = reshape(F,3,nNodes);
Freshape = Freshape';

for q = 1:allDoFs(end,1)
    node1  = allDoFs(q,2);
    p0     = [nodes(node1,2); nodes(node1,3)];
    p1     = [Freshape(node1,1); Freshape(node1,2)];
    quiver(p0(1), p0(2), p1(1), p1(2), .3, ...
            'MaxHeadSize', 1, 'LineWidth', 3, 'Color','b')
    hold on
end

axis equal
grid on

clearvars Index m n p0 p1 h q plotx ploty node1
